
(define-type FAE
  [num (n number?)]
  [add (lhs FAE?)
       (rhs FAE?)]
  [sub (lhs FAE?)
       (rhs FAE?)]
  [id (name symbol?)]
  [fun (param symbol?)
       (body FAE?)]
  [app (fun-expr FAE?)
       (arg-expr FAE?)]
  [if0 (test FAE?)
       (then FAE?)
       (else FAE?)])

(define-type FAE-Value
  [numV (n number?)]
  [closureV (param symbol?)
            (body FAE?)
            (ds DefrdSub?)])

(define-type DefrdSub
  [mtSub]
  [aSub (name symbol?)
        (value FAE-Value?)
        (rest DefrdSub?)])

;; ----------------------------------------

;; parse : S-expr -> FAE
(define (parse sexp)
  (cond
    [(number? sexp) (num sexp)]
    [(symbol? sexp) (id sexp)]
    [(pair? sexp)
     (case (car sexp)
       [(+) (add (parse (second sexp)) (parse (third sexp)))]
       [(-) (sub (parse (second sexp)) (parse (third sexp)))]
       [(fun) (fun (first (second sexp)) (parse (third sexp)))]
       [(if0) (if0 (parse (second sexp))
                   (parse (third sexp))
                   (parse (fourth sexp)))]
       [else (app (parse (first sexp)) (parse (second sexp)))])]))

(test (parse 3) (num 3))
(test (parse 'x) (id 'x))
(test (parse '{+ 1 2}) (add (num 1) (num 2)))
(test (parse '{- 1 2}) (sub (num 1) (num 2)))
(test (parse '{fun {x} x}) (fun 'x (id 'x)))
(test (parse '{1 2}) (app (num 1) (num 2)))
(test (parse '{if0 0 1 2}) (if0 (num 0) (num 1) (num 2)))

;; ----------------------------------------

;; interp : FAE DefrdSub (FAE-Value -> alpha) -> alpha
(define (interp a-fae ds k)
  (type-case FAE a-fae
    [num (n) (k (numV n))]
    [add (l r) (interp l ds
                       (lambda (v1)
                         (interp r ds
                                 (lambda (v2)
                                   (k (num+ v1 v2))))))]
    [sub (l r) (interp l ds
                       (lambda (v1)
                         (interp r ds
                                 (lambda (v2)
                                   (k (num- v1 v2))))))]
    [id (name) (k (lookup name ds))]
    [fun (param body-expr)
         (k (closureV param body-expr ds))]
    [app (fun-expr arg-expr)
         (interp fun-expr ds
                 (lambda (fun-val)
                   (interp arg-expr ds
                           (lambda (arg-val)
                             (interp (closureV-body fun-val)
                                     (aSub (closureV-param fun-val)
                                           arg-val
                                           (closureV-ds fun-val))
                                     k)))))]
    [if0 (test-expr then-expr else-expr)
         (interp test-expr ds
                 (lambda (v)
                   (if (numzero? v)
                       (interp then-expr ds k)
                       (interp else-expr ds k))))]))

;; num-op : (number number -> number) -> (FAE-Value FAE-Value -> FAE-Value)
(define (num-op op op-name)
  (lambda (x y)
    (numV (op (numV-n x) (numV-n y)))))

(define num+ (num-op + '+))
(define num- (num-op - '-))

;; numzero? : FAE-Value -> boolean
(define (numzero? x)
  (zero? (numV-n x)))

(define (lookup name ds)
  (type-case DefrdSub ds
    [mtSub () (error 'lookup "free variable")]
    [aSub (sub-name num rest-ds)
          (if (symbol=? sub-name name)
              num
              (lookup name rest-ds))]))


(test (interp (num 10)
              (mtSub)
              identity)
      (numV 10))
(test (interp (add (num 10) (num 7))
              (mtSub)
              identity)
      (numV 17))
(test (interp (sub (num 10) (num 7))
              (mtSub)
              identity)
      (numV 3))
(test (interp (app (fun 'x (add (id 'x) (num 12)))
                   (add (num 1) (num 17)))
              (mtSub)
              identity)
      (numV 30))
(test (interp (id 'x)
              (aSub 'x (numV 10) (mtSub))
              list)
      (list (numV 10)))

(test (interp (app (fun 'x (add (id 'x) (num 12)))
                   (add (num 1) (num 17)))
              (mtSub)
              identity)
      (numV 30))

(test (interp (if0 (num 0)
                   (num 1)
                   (num 2))
              (mtSub)
              identity)
      (numV 1))
(test (interp (if0 (num 1)
                   (num 0)
                   (num 2))
              (mtSub)
               identity)
      (numV 2))

(test (interp (parse 
               '{{fun {mkrec}
                      {{fun {fib}
                            ;; Call fib on 4:
                            {fib 4}}
                       ;; Create recursive fib:
                       {mkrec
                        {fun {fib}
                             ;; Fib:
                             {fun {n}
                                  {if0 n
                                       1
                                       {if0 {- n 1}
                                            1
                                            {+ {fib {- n 1}}
                                               {fib {- n 2}}}}}}}}}}
                 ;; mkrec:
                 {fun {body-proc}
                      {{fun {fX}
                            {fX fX}}
                       {fun {fX}
                            {body-proc {fun {x} {{fX fX} x}}}}}}})
              (mtSub)
              identity)
      (numV 5))
